/**
 * @namespace   biewlib
 * @file        biewlib/sysdep/ia32/win32/mmfio.c
 * @brief       This file contains implementation of memory mapped file i/o routines for Win32s.
 * @version     -
 * @remark      this source file is part of Binary vIEW project (BIEW).
 *              The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.
 *              All rights reserved. This software is redistributable under the
 *              licence given in the file "Licence.en" ("Licence.ru" in russian
 *              translation) distributed in the BIEW archive.
 * @note        Requires POSIX compatible development system
 *
 * @author      Nick Kurshev
 * @since       2000
 * @note        Development, fixes and improvements
**/
#ifdef __DISABLE_MMF
/*
   WARNING! Current implementation of program logic is not fully compatible
   with win32 MMF.
   WinNT 4.0 SDK says:
   SetEndOfFile.
    ...
   Remarks:
    ...
    If you called CreateFileMapping to create a file-mapping object for hFile,
    you must first call UnmapViewOfFile to unmap all views and call CloseHandle
    to close the file-mapping object before you can call SetEndOfFile.
*/
#include "biewlib/sysdep/ia16/dos/mmfio.c"
#else
#include <windows.h>
#include "biewlib/pmalloc.h"
#include "biewlib/biewlib.h"

struct mmfRecord
{
   void *   addr;
   long     length;
   int      fhandle;
   int      mode;
   HANDLE   fmapping;
};

static int __FASTCALL__ mk_prot(int mode)
{
  int pflg;
  pflg = PAGE_READONLY;
  if((mode & FO_WRITEONLY) || (mode & FO_READWRITE)) pflg = PAGE_READWRITE;
  pflg |= SEC_COMMIT | SEC_NOCACHE;
  if(mode & SO_PRIVATE) pflg |= PAGE_WRITECOPY;
  return pflg;
}

static int __FASTCALL__ mk_access(int mode)
{
  int pflg;
  pflg = FILE_MAP_READ;
  if((mode & FO_WRITEONLY) || (mode & FO_READWRITE)) pflg = FILE_MAP_WRITE;
  if(mode & SO_PRIVATE) pflg |= FILE_MAP_COPY;
  return pflg;
}

mmfHandle          __FASTCALL__ __mmfOpen(const char *fname,int mode)
{
  struct mmfRecord *mret;
  int fhandle;
  fhandle = __OsOpen(fname,mode);
  if(fhandle != -1)
  {
    mret = PMalloc(sizeof(struct mmfRecord));
    if(mret)
    {
      HANDLE fmapping;
      long length;
      void *addr;
      length = __FileLength(fhandle);
      fmapping = CreateFileMapping((HANDLE)fhandle,NULL,mk_prot(mode),0L,length,NULL);
      if(fmapping)
      {
        addr = MapViewOfFile(fmapping,mk_access(mode),0L,0L,length);
        if(addr)
        {
          mret->fhandle = fhandle;
          mret->fmapping= fmapping;
          mret->addr    = addr;
          mret->length  = length;
          mret->mode    = mode;
          return mret;
        }
        CloseHandle(fmapping);
      }
      PFree(mret);
    }
    __OsClose(fhandle);
  }
  return NULL;
}

tBool              __FASTCALL__ __mmfFlush(mmfHandle mh)
{
  struct mmfRecord *mrec = (struct mmfRecord *)mh;
  return FlushViewOfFile(mrec->addr,mrec->length) ? True : False;
}

mmfHandle     __FASTCALL__ __mmfSync(mmfHandle mh)
{
  /** @bug  This implementation does not change size of file.
            Sorry! It is not possible under win32 without changing logic of
            program.
  */
  struct mmfRecord *mrec = (struct mmfRecord *)mh;
  long length;
  length = __FileLength(mrec->fhandle);
  FlushViewOfFile(mrec->addr,min(mrec->length,length));
  UnmapViewOfFile(mrec->addr); /* It restores size of truncated file */
  CloseHandle(mrec->fmapping);
  if(length < mrec->length) __OsChSize(mrec->fhandle,length);
  mrec->length = __FileLength(mrec->fhandle);
  mrec->fmapping = CreateFileMapping((HANDLE)mrec->fhandle,NULL,mk_prot(mrec->mode),0L,mrec->length,NULL);
  if(mrec->fmapping)
  {
    if((mrec->addr = MapViewOfFile(mrec->fmapping,mk_access(mrec->mode),0L,0L,mrec->length)) != NULL)
                                                                    return mh;
    CloseHandle(mrec->fmapping);
  }
  __OsClose(mrec->fhandle);
  PFree(mrec);
  return NULL;
}

tBool              __FASTCALL__ __mmfProtect(mmfHandle mh,int flags)
{
  DWORD oldProt;
  struct mmfRecord *mrec = (struct mmfRecord *)mh;
  mrec->mode = flags;
  return VirtualProtect(mrec->addr,mrec->length,mk_prot(flags),&oldProt) ? True : False;
}

tBool              __FASTCALL__ __mmfResize(mmfHandle mh,long size)
{
  /** @bug  This implementation does not change size of file.
            Sorry! It is not possible under win32 without changing logic of
            program.
  */
  struct mmfRecord *mrec = (struct mmfRecord *)mh;
  UnmapViewOfFile(mrec->addr);
  CloseHandle(mrec->fmapping);
  __OsChSize(mrec->fhandle,size);
  mrec->length = __FileLength(mrec->fhandle);
  mrec->fmapping = CreateFileMapping((HANDLE)mrec->fhandle,NULL,mk_prot(mrec->mode),0L,mrec->length,NULL);
  if(mrec->fmapping)
  {
    if((mrec->addr = MapViewOfFile(mrec->fmapping,mk_access(mrec->mode),0L,0L,mrec->length)) != NULL)
                                                                    return True;
    CloseHandle(mrec->fmapping);
  }
  __OsClose(mrec->fhandle);
  PFree(mrec);
  return False;
}

void               __FASTCALL__ __mmfClose(mmfHandle mh)
{
  struct mmfRecord *mrec = (struct mmfRecord *)mh;
  UnmapViewOfFile(mrec->addr);
  CloseHandle(mrec->fmapping);
  __OsClose(mrec->fhandle);
  PFree(mrec);
}

void *             __FASTCALL__ __mmfAddress(mmfHandle mh)
{
  return ((struct mmfRecord *)mh)->addr;
}

long              __FASTCALL__ __mmfSize(mmfHandle mh)
{
  return ((struct mmfRecord *)mh)->length;
}

tBool             __FASTCALL__ __mmfIsWorkable( void ) { return True; }

#endif
